%{
This is a supplementary file for the paper of
Andrey Kan, Rajib Chakravorty, James Bailey,
Christopher Leckie, John Markham, Mark R. Dowling.

Andrey Kan (akan@csse.unimelb.edu.au) is the corresponding author.

Do not redistribute, do not use for commercial purposes.

Property of National ICT Australia.
=====================================

NENIA -- the NEarest Neighbour based Interframe Association

implementation of Hungarian algorithm from Markus Buehren
http://www.mathworks.com/matlabcentral/fileexchange/6543

detResults -- detection results, see "sampleDet.mat" for data format
cellData -- cell tracks, see "newCell.m" for data format
%}
function cellData = trackCellsNENIA(detResults)

trkPars.gd = analyseMovieMRD(detResults);
trkPars.gd2 = trkPars.gd ^2;

% main loop
videoMap = [];
labelsPrev = [];
ctx.lastTrackId	= 1;
ctx.currFrame	= 1;
endFrame        = size(detResults, 2);
while (ctx.currFrame <= endFrame)
	% do association
    locatsCurr = detResults(ctx.currFrame).centroid;
	[labelsCurr, ctx] = linkLocs(trkPars, ctx, labelsPrev, locatsCurr);
	videoMap = exportLabels(videoMap, labelsCurr, ctx.currFrame);

	% move to the next frame
	labelsPrev = labelsCurr;
	ctx.currFrame = ctx.currFrame + 1;
end

cellData = ctx.cellData;
end

function [nextLabels, ctx] = linkLocs(trkPars, ctx, labels, locations)

% main association round
labsNum = size(labels, 1);
locsNum = size(locations, 1);
distMatrix = zeros(labsNum, locsNum);
for labIdx = 1:labsNum
	lab = labels(labIdx);
	for locIdx = 1:locsNum
		loc = locations(locIdx,:);
		d2 = (loc(1,1) - lab.x)^2 + (loc(1,2) - lab.y)^2;
		if (d2 >= trkPars.gd2)
			d2 = Inf;
		end
		distMatrix(labIdx,locIdx) = d2;
	end
end
matching = assignmentoptimal(distMatrix);

% mantain active and lost tracks
usedMask = zeros(1, locsNum);
nextLabels = [];
for labIdx = 1:labsNum
	lab = labels(labIdx);
	if (numel(matching))
		locIdx = matching(labIdx);
	else
		locIdx = 0;
	end
	if (locIdx)
		% promote the current label
		loc = locations(locIdx,:);
		if (lab.id ~= 0)
			if (lab.splitFrame)
				ctx.cellData(lab.id).spltX = [ctx.cellData(lab.id).spltX  loc(1,1)];
				ctx.cellData(lab.id).spltY = [ctx.cellData(lab.id).spltY  loc(1,2)];
			else
				ctx.cellData(lab.id).locsX = [ctx.cellData(lab.id).locsX  loc(1,1)];
				ctx.cellData(lab.id).locsY = [ctx.cellData(lab.id).locsY  loc(1,2)];
			end
		end
		newLabel = promoteLabel(lab, loc);
		nextLabels = [nextLabels; newLabel]; %#ok<AGROW>
		usedMask(locIdx) = 1;
	else
		% not assigned
		if (~lab.lostFrame && ~lab.splitFrame)
			% active --> lost
			newLabel = lab;
			newLabel.prev.x = lab.x;
			newLabel.prev.y = lab.y;
			newLabel.lostFrame = ctx.currFrame;
			nextLabels = [nextLabels; newLabel]; %#ok<AGROW>
			if (lab.id ~= 0)
				ctx.cellData(lab.id).locsX ...
					= [ctx.cellData(lab.id).locsX  ctx.cellData(lab.id).locsX(end)];
				ctx.cellData(lab.id).locsY ...
					= [ctx.cellData(lab.id).locsY  ctx.cellData(lab.id).locsY(end)];
			end
		end
	end
end

% maintain splits
removeIdx = [];
labsNum = size(nextLabels, 1);
for labIdx = 1:labsNum
	lab = nextLabels(labIdx);
	if (lab.splitFrame && ~lab.lostFrame)
		% division candidate is active, find its origin
		isFound = 0;
		for lab0Idx = 1:labsNum
			lab0 = nextLabels(lab0Idx);
			if ((labIdx ~= lab0Idx) && (lab.id == lab0.id))
				if (~lab0.lostFrame)
					% pair is active, do split
					id1 = ctx.lastTrackId;
					id2 = ctx.lastTrackId + 1;
					ctx.lastTrackId = id2 + 1;
					
					ctx.cellData(id1) = newCell(ctx.currFrame-1, loc);
					ctx.cellData(id2) = newCell(ctx.currFrame-1, loc);
					ctx.cellData(id1).mother = lab.id;
					ctx.cellData(id2).mother = lab.id;
					ctx.cellData(id1).sister = id2;
					ctx.cellData(id2).sister = id1;
					
					if (lab0.splitMark == 'A')
						nextLabels(lab0Idx).id = id1; %#ok<AGROW>
						nextLabels(labIdx).id = id2; %#ok<AGROW>
						ctx.cellData(id1).locsX = ctx.cellData(lab.id).spltX;
						ctx.cellData(id1).locsY = ctx.cellData(lab.id).spltY;
						ctx.cellData(id2).locsX = ctx.cellData(lab.id).locsX(end-1:end);
						ctx.cellData(id2).locsY = ctx.cellData(lab.id).locsY(end-1:end);
					else
						nextLabels(labIdx).id = id1; %#ok<AGROW>
						nextLabels(lab0Idx).id = id2; %#ok<AGROW>
						ctx.cellData(id1).locsX = ctx.cellData(lab.id).locsX(end-1:end);
						ctx.cellData(id1).locsY = ctx.cellData(lab.id).locsY(end-1:end);
						ctx.cellData(id2).locsX = ctx.cellData(lab.id).spltX;
						ctx.cellData(id2).locsY = ctx.cellData(lab.id).spltY;
					end
					
					% parent
					ctx.cellData(lab.id).childL = id1;
					ctx.cellData(lab.id).childR = id2;
					ctx.cellData(lab.id).locsX(end-1:end) = [];
					ctx.cellData(lab.id).locsY(end-1:end) = [];
					ctx.cellData(lab.id).spltX = [];
					ctx.cellData(lab.id).spltY = [];
					
					% label, e.g. cell 3, divided, into 3A and 3B,
					% then 3AA and 3AB, and 3BA and 3BB, etc.
					nextLabels(lab0Idx).label = [lab0.label, lab0.splitMark]; %#ok<AGROW>
					nextLabels(labIdx).label = [lab0.label, lab.splitMark]; %#ok<AGROW>
					nextLabels(labIdx).splitFrame = 0; %#ok<AGROW>
				else
					% turn the candidate into the active original track
					nextLabels(labIdx).startFrame = lab0.startFrame; %#ok<AGROW>
					nextLabels(labIdx).splitFrame = 0; %#ok<AGROW>
					% remove the original track
					removeIdx = [removeIdx, lab0Idx]; %#ok<AGROW>
					
					ctx.cellData(lab.id).locsX(end-1:end) = ctx.cellData(lab.id).spltX;
					ctx.cellData(lab.id).locsY(end-1:end) = ctx.cellData(lab.id).spltY;
				end
				isFound = 1;
				break;
			end
		end
		if (isFound == 0)
			error('split pair not found');
		end
	end
end
if (size(removeIdx, 2) > 0)
	nextLabels(removeIdx) = [];
end

% second association round: try splits
labsNum = size(nextLabels, 1);
distMatrix = zeros(labsNum, locsNum);
for labIdx = 1:labsNum
	lab = nextLabels(labIdx);
	if (lab.splitFrame)
		error('split candidates must be resolved by this line');
	end
	if (ctx.currFrame - lab.startFrame <= 1 && lab.startFrame ~= 1)
		% don't try to do splits from fresh splits
		distMatrix(labIdx,:) = Inf;
		continue
	end
	if ((lab.prev.x == 0) && (lab.prev.y == 0))
		error('there is always the previous point');
	else
		point.x = lab.prev.x;
		point.y = lab.prev.y;
	end
	for locIdx = 1:locsNum
		distMatrix(labIdx,locIdx) = Inf;
		if (usedMask(locIdx) == 0)
			% work with previously unassigned observations
			loc = locations(locIdx,:);
			d2 = (loc(1,1) - point.x)^2 + (loc(1,2) - point.y)^2;
			if (d2 < trkPars.gd2)
				if (lab.lostFrame)
					fprintf('frame %d\n', ctx.currFrame);
					fprintf('lab.id %d\n', lab.id);
					fprintf('locIdx %d\n', locIdx);
					error('this observation must have been assigned in the first round');
				end
				distMatrix(labIdx,locIdx) = d2;
			end
		end
	end
end
matching = assignmentoptimal(distMatrix);

% second association round: create split candidates
for labIdx = 1:labsNum
	if (numel(matching))
		locIdx = matching(labIdx);
	else
		locIdx = 0;
	end
	if (locIdx)
		lab = nextLabels(labIdx);
		loc = locations(locIdx,:);
		newLab = splitLabel(ctx, lab, loc);
		ctx.cellData(lab.id).spltX = loc(1,1);
		ctx.cellData(lab.id).spltY = loc(1,2);
		oldMark = 'A';
		newMark = 'B';
		if ((newLab.x < lab.x) || ((newLab.x == lab.x) && (newLab.y < lab.y)))
			oldMark = 'B';
			newMark = 'A';
		end
		nextLabels(labIdx).splitMark = oldMark; %#ok<AGROW>
		newLab.splitMark = newMark;
		nextLabels = [nextLabels; newLab]; %#ok<AGROW>
		usedMask(locIdx) = 1;
	end
end

% initialize new tracks
for locIdx = 1:locsNum
	if (usedMask(locIdx) == 0)
		loc = locations(locIdx,:);
		ctx.cellData(ctx.lastTrackId) = newCell(ctx.currFrame, loc);
		nextLabels = [nextLabels; initLabel(loc, ctx.lastTrackId, ctx.currFrame)]; %#ok<AGROW>
		ctx.lastTrackId = ctx.lastTrackId + 1;
	end
end
end

function newLabel = promoteLabel(label, location)
% promote label
newLabel.id		= label.id;
newLabel.label	= label.label;
newLabel.x		= location(1,1);
newLabel.y		= location(1,2);
newLabel.prev.x = label.x;
newLabel.prev.y = label.y;
newLabel.startFrame = label.startFrame;
newLabel.lostFrame	= 0;
newLabel.splitFrame = label.splitFrame;
newLabel.splitMark	= label.splitMark;
end

function newLabel = splitLabel(ctx, label, location)
newLabel.id = label.id;
newLabel.label = label.label;
newLabel.x = location(1,1);
newLabel.y = location(1,2);
newLabel.prev.x = label.prev.x;
newLabel.prev.y = label.prev.y;
newLabel.startFrame = ctx.currFrame;
newLabel.lostFrame = 0;
newLabel.splitFrame = ctx.currFrame;
newLabel.splitMark = 'B';
end

function newLabel = initLabel(location, trackId, startFrame)
newLabel.id = trackId;
newLabel.label = num2str(newLabel.id);
newLabel.x = location(1,1);
newLabel.y = location(1,2);
newLabel.prev.x = 0;
newLabel.prev.y = 0;
newLabel.startFrame = startFrame;
newLabel.lostFrame = 0;
newLabel.splitFrame = 0;
newLabel.splitMark = 'A';
end

function videoMap = exportLabels(videoMap, labels, currFrame)
frameMap.num = currFrame;
frameMap.map = [];

labNum = size(labels, 1);
for labIdx = 1:labNum
	lab = labels(labIdx);
	if (~lab.lostFrame)
		pointMap.x = lab.x;
		pointMap.y = lab.y;
		pointMap.id = lab.id;
		pointMap.label = lab.label;
		frameMap.map = [frameMap.map; pointMap];
	end
end

end
